@@ -5,10 +5,13 @@
 
 #include "arch.h"
 
-#define MG_DEVICE_NONE 0      // Dummy system
-#define MG_DEVICE_STM32H5 1   // STM32 H5
-#define MG_DEVICE_STM32H7 2   // STM32 H7
-#define MG_DEVICE_CUSTOM 100  // Custom implementation
+#define MG_DEVICE_NONE 0        // Dummy system
+#define MG_DEVICE_STM32H5 1     // STM32 H5
+#define MG_DEVICE_STM32H7 2     // STM32 H7
+#define MG_DEVICE_RT1020 3      // IMXRT1020
+#define MG_DEVICE_RT1060 4      // IMXRT1060
+#define MG_DEVICE_CH32V307 100  // WCH CH32V307
+#define MG_DEVICE_CUSTOM 1000   // Custom implementation
 
 #ifndef MG_DEVICE
 #define MG_DEVICE MG_DEVICE_NONE
@@ -23,7 +26,7 @@ int mg_flash_bank(void);            // 0: not dual bank, 1: bank1, 2: bank2
 
 // Write, erase, swap bank
 bool mg_flash_write(void *addr, const void *buf, size_t len);
-bool mg_flash_erase(void *addr);  // Must be at sector boundary
+bool mg_flash_erase(void *sector);
 bool mg_flash_swap_bank(void);
 
 // Convenience functions to store data on a flash sector with wear levelling
